@@ -47,7 +47,7 @@ def reload_eo(filename):
     dict : evolutionary optimization data
     
     """
-    if type(filename) == str:
+    if type(filename) in [str, np.str_]:
         try:
             eo_dict = pickle.load( open(filename, "rb") )
         except IOError:
@@ -59,6 +59,10 @@ def reload_eo(filename):
         
     return eo_dict
     
+def load_eo(filename):
+    """wrapper for `reload_eo`"""
+    return reload_eo(filename)
+    
 
 
 #==============================================================================
