@@ -1,6 +1,6 @@
 # encoding: utf-8
 #
-#Copyright (C) 2017-2021, P. R. Wiecha
+#Copyright (C) 2017-2022, P. R. Wiecha
 #
 #This program is free software: you can redistribute it and/or modify
 #it under the terms of the GNU General Public License as published by
@@ -19,22 +19,30 @@
     pyGDM2 - a full-field electrodynamical solver python toolkit.
     Based on the Green Dyadic Method.
     
+    Contributors:
+        - P.R. Wiecha (maintainer)
+        - Ch. Girard
+        - A. Arbouet
+        - C. Majorel
+        - A. Patoux
+        - G. Colas des Francs
+        - Y. Brûlé
+    
 """
-from pyGDM2 import core as core_py
-from pyGDM2 import fields as fields_py
-from pyGDM2 import linear as linear_py
-
-
 __name__ = 'pyGDM2'
-__version__ = '1.1.2'
-__date__ = "12/10/2021"   # MM/DD/YYY
+__version__ = '1.1.3'
+__date__ = "04/26/2022"   # MM/DD/YYY
 __author__ = 'Peter R. Wiecha'
 
 __all__ = ["core", "materials", "structures", "fields", 
            "propagators", "propagators_2D",
            "linear", "nonlinear", 
-           "tools", "visu",
-           "core_py", "fields_py", "linear_py", 
-           "EO", "EO.core", "EO.tools", "EO.models", "EO.problem"]
+           "tools", "visu", 
+           "EO"]
 
 
+## compatibility with deprecated transition modules:
+from pyGDM2 import core as core_py
+from pyGDM2 import fields as fields_py
+from pyGDM2 import linear as linear_py
+
