@@ -1,6 +1,6 @@
 # encoding: utf-8
 #
-#Copyright (C) 2017-2021, P. R. Wiecha
+#Copyright (C) 2017-2022, P. R. Wiecha
 #
 #This program is free software: you can redistribute it and/or modify
 #it under the terms of the GNU General Public License as published by
@@ -31,7 +31,7 @@ import time
 
 import numba
 
-from pyGDM2 import propagators
+from pyGDM2.propagators import propagators
 
 
 
@@ -288,10 +288,6 @@ def scatter(sim, method='lu', calc_H=False, verbose=True, callback=None, **kwarg
 # =============================================================================
     if method.lower() == "cuda":
         method = "cupy"
-    if method.lower() == "cupy":
-        import cupy
-        if int(cupy.__version__[0])<7:
-            raise ValueError("`cupy` version 7 or higher required. Found cupy version {}. Please upgrade cupy.".format(cupy.__version__))
         
     if method.lower() not in ["numpyinv", "scipyinv", "lu", "cupy"]:
         raise ValueError('Error: Unknown solving method. Must be one of' +
@@ -399,6 +395,130 @@ def scatter(sim, method='lu', calc_H=False, verbose=True, callback=None, **kwarg
 
 
 
+## !!! Not working due to numba conflicting with multiprocessing
+def _scatter_multiprocessing(sim, calc_H=False, verbose=False, scatter_verbose=False, 
+                            N_process=-1, N_thread=1, **kwargs):
+    """multiprocessing parallel execution of :func:`.scatter` for parallel calculation of spectra
+    
+    *requires:* **threadpoolctl** (for setting number of numpy parallel threads)
+    
+    Parameters
+    ----------
+    sim : :class:`.simulation` 
+        simulation description
+    
+    calc_H : bool, default: False
+        if True, calculate also the internal magnetic field H.
+    
+    verbose, scatter_verbose: bool, default: False
+        turns on some mpi-routine info printing, respectively controls verbose setting for :func:`.scatter`
+    
+    N_process : int, default: -1
+        optional number of parallel processes to use for scattering evaluation. 
+        '-1': use number of available CPUs
+    
+    N_thread : int, default: 1
+        optional number of CPUS to allocate for each thread (each wavelength). 
+        '-1': use number of available CPUs
+    
+    **kwargs : 
+        all kwargs are passed to :func:`.scatter`
+
+
+    Notes 
+    -----
+    - The product of `N_process` x `N_thread` shoould not exceed the number of CPUs.
+      *Overloading the CPUs will usually result in **decreased** calculation speed!*
+    
+    - The memory requirements are `N_process` times higher as for single execution!
+    
+    - see :func:`.scatter` for main documentation
+    
+    """
+    from pyGDM2 import fields
+    from pyGDM2.tools import _get_nr_processes
+    import multiprocessing
+    
+    if N_process == -1:
+        N_process = _get_nr_processes()
+    if N_thread == -1:
+        N_thread = _get_nr_processes()
+    if verbose: 
+        print("")
+        print("number of processes:             ", N_process)
+        print("number of threads / wavelength:  ", N_thread)
+        print("number of wavelengths / process: ", int(np.ceil(len(sim.efield.wavelengths) / float(N_process))))
+        print("")
+    
+    from threadpoolctl import threadpool_limits
+    import numba
+    threadpool_limits(limits=N_thread, user_api='blas')
+    numba.set_num_threads(N_thread)
+    
+    
+    
+    ## --- create list of jobs and split in equal parts depending on `nprocs`
+    def split(jobs, nprocs):
+        return [jobs[i::nprocs] for i in range(nprocs)]
+    
+    if verbose: print("Generating jobs... ", end='')
+    jobs = []
+    for i, wl in enumerate(sim.efield.wavelengths):
+        ## --- generate simulation objects for each individual wavelength
+        _sim = copy.deepcopy(sim)
+        _efield = fields.efield(_sim.efield.field_generator, [wl], _sim.efield.kwargs)
+        _efield.setDtype(_sim.dtypef, _sim.dtypec)
+        _sim.efield = copy.deepcopy(_efield)
+        jobs.append(_sim)
+    # jobs = split(jobs_all, N_process)
+    if len(jobs)%N_process != 0:
+        warnings.warn("Efficiency warning: Number of wavelengths ({}) ".format(
+                                            len(sim.efield.wavelengths)) + 
+                      "not divisable by Nr of processes ({})!".format(N_process))
+    if verbose: print("Done.")
+    
+    
+    ## --- evaluation of `scatter` in parallel for each wavelength
+    if verbose: print("Running on {} processes... ".format(N_process), end='')
+    t0 = time.time()
+    global worker_wrapper
+    def worker_wrapper(job):
+        return scatter(job, calc_H=calc_H, verbose=scatter_verbose, **kwargs)
+    
+    ## --- multiprocessing on N_cpu tasks
+    with multiprocessing.Pool(N_process) as p:
+        _ = p.map(worker_wrapper, jobs)
+    
+    ## --- gathering results
+    if verbose: print("All simulations done. Recombining... ", end='')
+    resultsE = []
+    resultsH = []
+    for job in jobs:
+        scattered_Efield = job.E
+        for _scatf in scattered_Efield:
+            resultsE.append(_scatf)
+        if calc_H:
+            scattered_Hfield = job.H
+            for _scatf in scattered_Hfield:
+                resultsH.append(_scatf)
+    
+    ## --- recombine data and sort by wavelength
+    resultsE = [i for temp in resultsE for i in temp]
+    resultsE = sorted(resultsE, key=lambda k: k[0]['wavelength'])
+    sim.E = resultsE
+    if calc_H:
+        resultsH = [i for temp in resultsH for i in temp]
+        resultsH = sorted(resultsH, key=lambda k: k[0]['wavelength'])
+        sim.H = resultsH
+    if verbose: print("Done.")
+    
+    
+    if verbose:
+        print('calculated {} wavelengths in {:.2f}s (working on {} processes)'.format(
+                        len(sim.efield.wavelengths), time.time()-t0, N_process))
+    
+    return 1
+    
         
 def scatter_mpi(sim, calc_H=False, verbose=False, scatter_verbose=False, **kwargs):
     """MPI wrapper to :func:`.scatter` for embarrassingly parallel calculation of spectra
@@ -428,15 +548,15 @@ def scatter_mpi(sim, calc_H=False, verbose=False, scatter_verbose=False, **kwarg
     -----
     - On single machines it is usually easier to install `scipy` 
       compiled with parallel BLAS (parallel LU / CG routines). Usually the
-      parallel BLAS is will be already installed automatically. Try to not use 
-      both parallelisation techniques simultaneously unless properly
-      configured for instance via a batch script (e.g. `SLURM`). 
-      *Overloading the CPUs will usually result in decreased calculation speed.*
+      parallel BLAS is installed by default. Using both
+      parallelisation techniques simultaneously requires proper
+      configuration of threads / process numbers. 
+      *Overloading the CPUs will usually result in **decreased** calculation speed!*
     
     - see :func:`.scatter` for main documentation
     
     """
-    from pyGDM2 import fields_py as fields  # !!! adapt namespace in final version
+    from pyGDM2 import fields
     from mpi4py import MPI
     
     comm = MPI.COMM_WORLD
@@ -466,7 +586,6 @@ def scatter_mpi(sim, calc_H=False, verbose=False, scatter_verbose=False, **kwarg
             _sim = copy.deepcopy(sim)
             _efield = fields.efield(_sim.efield.field_generator, [wl], _sim.efield.kwargs)
             _efield.setDtype(_sim.dtypef, _sim.dtypec)
-#            _sim.efield.wavelengths = [wl]
             _sim.efield = copy.deepcopy(_efield)
             jobs_all.append(_sim)
         jobs = split(jobs_all, nprocs)
@@ -660,6 +779,9 @@ def decay_rate(sim, field_index=None, wavelength=None,
 # =============================================================================
 # preparation
 # =============================================================================
+    r_probe = np.array(r_probe)
+    if len(r_probe.shape) == 1:
+        r_probe = r_probe[None,:]
     if r_probe.shape[0]==3 and r_probe.shape[1]!=3:
         r_probe = np.transpose(r_probe)
     
@@ -861,6 +983,7 @@ def _get_free_GPU_memory():
 
 
 def get_general_propagator(sim=None, wavelength=None, method='lu', 
+                           fallbackmethod_gpu='lu',
                            calc_H=False, verbose=False):
     """invert dipole-coupling matrix
     
@@ -880,6 +1003,9 @@ def get_general_propagator(sim=None, wavelength=None, method='lu',
          - "cupy" uses CUDA GPU via `cupy`
          - "cuda" (equivalent to "cupy")
     
+    fallbackmethod_gpu : string, default: "lu"
+        method to fall back to if using cuda solver and GPU runs out of RAM 
+    
     calc_H : bool, default: False
         if True, calculate also the internal magnetic field H. This has an 
         impact on performance and memory requirement, so it is deactivated by 
@@ -917,7 +1043,7 @@ def get_general_propagator(sim=None, wavelength=None, method='lu',
         
     if method.lower() == "cupy":
         import cupy
-        if int(cupy.__version__[0])<7:
+        if int(cupy.__version__.split('.')[0])<7:
             raise ValueError("`cupy` version 7 or higher required. Found cupy " +
                              "version {}. Please upgrade cupy.".format(cupy.__version__))
     
@@ -957,7 +1083,7 @@ def get_general_propagator(sim=None, wavelength=None, method='lu',
             req_mem = 4 * M_EE.nbytes/1024**2    # 'inv' requires memorey of 4*array
             if req_mem > free_mem:
                 warnings.warn("Required memory exceeds GPU RAM. Falling back to scipy's 'scipyinv'. You could also try 'LU' instead, which is often more efficient.")
-                method = 'scipyinv'
+                method = fallbackmethod_gpu
             else:
                 ## move array to cuda device, cuda inversion
                 Agpu = cp.array(np.ascontiguousarray(M_EE))
