@@ -1,6 +1,6 @@
 # encoding: utf-8
 #
-#Copyright (C) 2017-2021, P. R. Wiecha
+#Copyright (C) 2017-2022, P. R. Wiecha
 #
 #This program is free software: you can redistribute it and/or modify
 #it under the terms of the GNU General Public License as published by
@@ -107,29 +107,62 @@ def save_simulation(sim, fname, mode='h5',
     if mode.lower() in ['h5', 'hdf5']:
         try:
             import tables
+            sim_save = sim.copy()
+            FILTERS = tables.Filters(**tables_filter_kwargs)
+            f = tables.open_file(fname + '.h5', 'w', filters=FILTERS)  # create h5 file
             
-                    ## write E-field to h5
+            
+            ## --- write structure mesh to h5
+            grp_geo = f.create_group('/', 'struct')
+            f.create_array(grp_geo, 'geometry', obj=sim.struct.geometry)
+            f.create_array(grp_geo, 'step', obj=sim.struct.step)
+                
+            ## --- write E-field to h5
             if sim.E is not None:
-                FILTERS = tables.Filters(**tables_filter_kwargs)
-                f = tables.open_file(fname + '.h5', 'w', filters=FILTERS)
-                
                 grp_E = f.create_group('/', 'E')
-                
                 E_dicts = []
                 for i, ef in enumerate(sim.E):
                     Edict = ef[0]
                     E_dicts.append([Edict, None])   # remove E-field-data
                     E = ef[1]
                     f.create_carray(grp_E, 'E{}'.format(i), obj=E)
-                f.close()
                 
-                ## write rest (without sim.E fields) to pickle file
-                E = sim.E
-                sim.E = E_dicts    
-                pickle_sim(sim)
-                sim.E = E
-            else:
-                pickle_sim(sim)
+                ## remove E fields from pickle-part of sim
+                sim_save.E = E_dicts    
+            
+            ## --- write H-field to h5
+            if hasattr(sim, 'H'):
+                if sim.H is not None:
+                    grp_H = f.create_group('/', 'H')
+                    H_dicts = []
+                    for i, hf in enumerate(sim.H):
+                        Hdict = hf[0]
+                        H_dicts.append([Hdict, None])   # remove E-field-data
+                        H = hf[1]
+                        f.create_carray(grp_H, 'H{}'.format(i), obj=H)
+                    
+                    ## remove H fields from pickle-part of sim
+                    sim_save.H = H_dicts    
+            
+            ## --- write generalized polarizabilities to h5
+            def generalized_pola_to_h5(_sim, grp_name):
+                if hasattr(_sim, grp_name):
+                    g_p = getattr(_sim, grp_name)
+                    grp_gp = f.create_group('/', grp_name)
+                    for i_k, wl_key in enumerate(g_p.keys()):
+                        f.create_carray(grp_gp, 'i{}'.format(i_k), obj=g_p[wl_key], title=str(wl_key))
+                    
+                        ## remove general. polas from pickle-part of sim
+                        g_p[wl_key] = None
+                        # getattr(_sim, grp_name, None)
+            
+            for gp_name in ['K_P_E', 'K_T_E', 'K_QE_E', 'K_QT_E', 'K_M_E', 'K_QM_E']:
+                generalized_pola_to_h5(sim_save, gp_name)
+                
+            ## --- save rest (config)
+            pickle_sim(sim_save)
+            
+            f.close()    # close h5 file
             
         except ImportError:
             warnings.warn("'hdf5' mode requires `tables` module. Seems not to be installed. " + 
@@ -162,10 +195,12 @@ def load_simulation(fname, pickle_encoding=None):
     -------
     instance of :class:`.core.simulation`
     """
-    ## check extension, if none, use '.pygdmsim'
+    ## check extension, if none, try using '.pygdmsim'
     import os
-    if os.path.splitext(fname)[1] == '':
+    if os.path.splitext(fname)[1] == '' and not os.path.isfile(fname):
         fname += '.pygdmsim'
+    if not os.path.isfile(fname):
+        raise Exception('File not found!')
         
         
     ## --- try first hdf5, if fails, use pure pickle
@@ -182,12 +217,31 @@ def load_simulation(fname, pickle_encoding=None):
         else:
             sim = pickle.load( open( fname, "rb" ), encoding=pickle_encoding )
         
-        ### efields from hdf5
+        ## --- read E fields from hdf5
         f = tables.open_file(fname + '.h5', 'r')
-        for ef in f.root.E:
-            enr = int(ef.name[1:])
-            E = ef.read()
-            sim.E[enr][1] = E
+        if "/E" in f:
+            for ef in f.root.E:
+                enr = int(ef.name[1:])
+                E = ef.read()
+                sim.E[enr][1] = E
+        
+        ## --- read H fields from hdf5
+        if "/H" in f:
+            for hf in f.root.H:
+                hnr = int(hf.name[1:])
+                H = hf.read()
+                sim.H[hnr][1] = H
+        
+        ## --- read generalized polarizabilties from hdf5
+        for gp_name in ['K_P_E', 'K_T_E', 'K_QE_E', 'K_QT_E', 'K_M_E', 'K_QM_E']:
+            if hasattr(sim, gp_name) and "/"+gp_name in f:
+                h5group = getattr(f.root, gp_name)
+                for _GP in h5group:
+                    gp_key = float(_GP.title)
+                    _GP_dat = _GP.read()
+                    ## get corresponding dict and set wavelength element
+                    getattr(sim, gp_name)[gp_key] = _GP_dat
+            
         f.close()
         
     except (IOError, ImportError) as e:
@@ -253,14 +307,13 @@ def print_sim_info(sim, prnt=True, verbose=False):
     
     >>> print_sim_info(sim)
     
-    will result in the identical output
+    will result in identical output
     
     """
     if not hasattr(sim, 'dyads'):   # ! COMPATIBILITY WITH OLD API
         return _print_sim_info_old_API(sim, prnt, verbose)
     else:
         
-        
         out_str = ''
         out_str += '\n' + ' =============== GDM Simulation Information ==============='
         out_str += '\n' + 'numerical precision: {} / {}'.format(sim.dtypef, sim.dtypec)
@@ -284,14 +337,20 @@ def print_sim_info(sim, prnt=True, verbose=False):
             out_str += '\n' + bcolors.FAIL + '   NO self-consistent E-fields'.format() + bcolors.ENDC
         else:
             out_str += '\n' + bcolors.OKGREEN + '   self-consistent E-fields are available'.format() + bcolors.ENDC
+        
         try:
             if sim.H is None:
                 out_str += '\n' + bcolors.FAIL + '   NO self-consistent H-fields'.format() + bcolors.ENDC
             else:
                 out_str += '\n' + bcolors.OKGREEN + '   self-consistent H-fields are available'.format() + bcolors.ENDC
-        except NameError:
+        except AttributeError:
             pass
         
+        if hasattr(sim, "K_P_E"):
+            out_str += '\n' + ''
+            out_str += '\n' + ' ===== *multipole* ======'
+            out_str += '\n' + bcolors.OKGREEN + '   Generalized polarizabilties available'.format() + bcolors.ENDC
+    
         if prnt:
             print(out_str)
         else:
@@ -864,7 +923,7 @@ def get_step_from_geometry(struct, max_meshpoints=1000):
     return step
 
 
-def get_mesh_from_geometry(geo, N_try=20, verbose=0):
+def get_mesh_from_geometry(geo, N_try=20, warn=True, verbose=0):
     """Try to identify the mesh type from the dipole positions
     
     If meshpoint determination fails, falling back to `cube`
@@ -879,6 +938,9 @@ def get_mesh_from_geometry(geo, N_try=20, verbose=0):
         maximum number of random evaluation point tries in order to find a
         "bulk" meshpoint.
         
+    warn: bool, default: True
+        show warnings if detection fails
+        
     verbose: int, default: 0
         if 1, print runtime info
         
@@ -891,6 +953,7 @@ def get_mesh_from_geometry(geo, N_try=20, verbose=0):
     lookup_meshtype = {
         10.24:"cube",  # 3D
         10.66:"cube",  # 2-layer 3D
+        11.07:"cube",  # 5-next neighbors 3D
         11.66:"cube",  # 2D
 
         9.0:"hex",     # 3D
@@ -920,7 +983,8 @@ def get_mesh_from_geometry(geo, N_try=20, verbose=0):
         if verbose:
             print("step= {:.1f}, distsum={:.2f} --> mesh: '{}'".format(step, distsum, meshtype))
     except KeyError:
-        warnings.warn("Mesh not detected, falling back to 'cubic'. mindist value = {}".format(distsum))
+        if warn:
+            warnings.warn("Mesh not detected, falling back to 'cubic'. mindist value = {}".format(distsum))
         meshtype = 'cube'
     
     return meshtype
@@ -1141,9 +1205,6 @@ def get_closest_field_index(sim, search_kwargs):
     index : int
         index of field which matches closest search incident field parameters
     """
-    if sim.E is None: 
-        warnings.warn("Scattered field inside the structure not yet evaluated. Run the `core.scatter` simulation before further evaluations.")
-    
     ## -- create dict of differences compared to search-parameters
     keys = get_field_indices(sim)
     
@@ -1188,9 +1249,13 @@ def get_possible_field_params_spectra(sim):
     return sorted(sim.efield.kwargs_permutations, key=sorted)
 
 
-def calculate_spectrum(sim, field_kwargs, func, verbose=False, callback=None, **kwargs):
+def calculate_spectrum_sequential(sim, field_kwargs, func, verbose=False, callback=None, **kwargs):
     """calculate spectra using function `func`
     
+    * ----------- Deprecated!!! ----------- * 
+    Old implementation of purely sequential spectrum calculation.
+    Will be removed in future version!!
+    
     Parameters
     ----------
     sim : :class:`.simulation`
@@ -1269,6 +1334,150 @@ def calculate_spectrum(sim, field_kwargs, func, verbose=False, callback=None, **
         pass
     
     return wl, spectrum
+
+
+
+
+
+def _get_nr_processes():
+    """return available processes
+    
+    see: 
+    https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python
+    """
+    ## preffered method to get available processes (might fail on windows)
+    try:
+        import os
+        return len(os.sched_getaffinity(0))
+    except:
+        pass
+    
+    ## if failed, try alternative using psutils
+    try:
+        import psutil
+        return len(psutil.Process().cpu_affinity())
+    except:
+        pass
+    
+    ## fall back on multiprocessing value (if psutils not installed)
+    import multiprocessing
+    return multiprocessing.cpu_count()
+
+
+def calculate_spectrum(sim, field_kwargs, func, verbose=False, callback=None, 
+                       N_cpu=1, **kwargs):
+    """calculate spectra using function `func`
+    
+    Parameters
+    ----------
+    sim : :class:`.simulation`
+        simulation description
+    
+    field_kwargs : `dict` or `int`
+        dict of field-configuration to use from simulation. If an `int` is given,
+        it is used as field-index (use e.g. :func:`get_closest_field_index`)
+    
+    func : function
+        evaluation function, which will be called as:
+        `func(sim, field_index, **kwargs)`
+    
+    callback : func, default: None
+        *only has effect in sequential mode (`N_cpu=1`)*
+        optional callback function, which is called after each wavelength.
+        Passes a dict to `callback` containing current wavelength, so far
+        calculated spectrum and timing info.
+    
+    N_cpu : int, default: 1
+        optional number of parallel tasks to use for function evaluation. 
+        '-1': use number of available CPUs
+    
+    **kwargs : 
+        all other keyword args are passed to `func`
+    
+    Returns
+    -------
+    np.array : wavelengths
+    
+    np.array : results of `func` for each wavelength
+    
+    """
+    import time
+    import multiprocessing
+    if N_cpu == -1:
+        N_cpu = _get_nr_processes()
+    
+    
+    if type(field_kwargs) == int:
+        kw_index = field_kwargs
+        field_kwargs = get_possible_field_params_spectra(sim)[field_kwargs]
+        if verbose:
+            warnings.warn("'field_kwargs' are index instead of dict." + 
+                      "Using `get_possible_field_params_spectra`." + 
+                      "Using configuration #{}: '{}'".format(kw_index, field_kwargs))
+        
+    
+    if len(field_kwargs) != len(get_field_indices(sim)[0])-1 or 'wavelength' in field_kwargs:
+        raise ValueError("'field_kwargs' must define every field-parameter except 'wavelength'!")
+    
+    ## --- determine field-indices for spectrum
+    wl = []
+    fidx_list_spec = []
+    for i, params in enumerate(get_field_indices(sim)):
+        # params = E[0]
+        skip = False
+        for kw in field_kwargs:
+            try:
+                if params[kw] != field_kwargs[kw]:
+                    skip = True
+            except ValueError:
+                if np.all(params[kw] != field_kwargs[kw]):
+                    skip = True
+        if not skip:
+            wl.append(params['wavelength'])
+            fidx_list_spec.append(i)
+            
+    
+    ## --- calculate the spectrum 
+    global worker_wrapper
+    def worker_wrapper(fidx):
+        return func(sim, fidx, **kwargs)
+    
+    t0 = time.time()
+    if N_cpu == 1:
+        ## --- sequentiel evaluation (non-multiprocessing)
+        spectrum = []
+        for fidx in fidx_list_spec:
+            t1 = time.time()
+            dat = worker_wrapper(fidx)
+            spectrum.append(dat)
+            if callback is not None:
+                cb_continue = callback(dict(
+                    spectrum=spectrum, wavelengths=wl,
+                    t_wl=1000.*(time.time()-t1)
+                                  ))
+                if not cb_continue:
+                    break           # quit if callback returns False
+    else:
+        ## --- multiprocessing on N_cpu tasks
+        t0 = time.time()
+        with multiprocessing.Pool(N_cpu) as p:
+            spectrum = p.map(worker_wrapper, fidx_list_spec)
+    
+    if verbose:
+        print('calculated {} wavelengths in {:.2f}s (working on {} processes)'.format(
+                                    len(wl), time.time()-t0, N_cpu))
+    
+    
+    ## --- try to convert to numpy arrays
+    try:
+        wl = np.array(wl)
+        spectrum = np.array(spectrum)
+        if len(spectrum) > 1 and spectrum.shape[-1] == 1:
+            spectrum = spectrum.T[0].T
+    except:
+        pass
+    
+    return wl, spectrum
     
 
     
@@ -1799,7 +2008,10 @@ def get_intensity_from_fieldlist(NF_list):
 #==============================================================================
 # generate 2D map coordinate lists
 #==============================================================================
-def generate_NF_map(X0,X1,NX,  Y0,Y1,NY,  Z0=0, projection='XY', dtype='f'):
+def generate_coord_map_planar(X0,X1,NX, Y0,Y1,NY, Z0=0, projection='XY', dtype='f'):
+    """wrapper to :func:`generate_NF_map`"""
+    return generate_NF_map(X0,X1,NX, Y0,Y1,NY, Z0, projection, dtype)
+def generate_NF_map(X0,X1,NX, Y0,Y1,NY, Z0=0, projection='XY', dtype='f'):
     """Generate coordinate list with equidistant positions on a rectangular map
     
     e.g. for nearfield calcuation on this map
@@ -1836,18 +2048,21 @@ def generate_NF_map(X0,X1,NX,  Y0,Y1,NY,  Z0=0, projection='XY', dtype='f'):
       
     """
     if projection.lower() =='xy':
-        MAP = generate_NF_map_XY(X0,X1,NX,  Y0,Y1,NY,  Z0, dtype)
+        r_probe = generate_NF_map_XY(X0,X1,NX, Y0,Y1,NY, Z0, dtype)
     elif projection.lower() =='xz':
-        MAP = generate_NF_map_XZ(X0,X1,NX,  Y0,Y1,NY,  Z0, dtype)
+        r_probe = generate_NF_map_XZ(X0,X1,NX, Y0,Y1,NY, Z0, dtype)
     elif projection.lower() == 'yz':
-        MAP = generate_NF_map_YZ(X0,X1,NX,  Y0,Y1,NY,  Z0, dtype)
+        r_probe = generate_NF_map_YZ(X0,X1,NX, Y0,Y1,NY, Z0, dtype)
     else:
         raise ValueError("Invalid projection specified. Must be one of: ['XY','XZ','YZ'].")
     
-    return MAP
+    return r_probe
 
 
-def generate_NF_map_XY(X0,X1,NX,  Y0,Y1,NY,  Z0=0, dtype='f'):
+def generate_coord_map_XY(X0,X1,NX, Y0,Y1,NY, Z0=0, dtype=None):
+    """wrapper to :func:`generate_NF_map_XY`"""
+    return generate_NF_map_XY(X0,X1,NX, Y0,Y1,NY, Z0, dtype)
+def generate_NF_map_XY(X0,X1,NX, Y0,Y1,NY, Z0=0, dtype=None):
     """Generate coordinate list with equidistant positions on a rectangular map in XY plane
     
     see :func:`generate_NF_map` for doc.
@@ -1858,15 +2073,15 @@ def generate_NF_map_XY(X0,X1,NX,  Y0,Y1,NY,  Z0=0, dtype='f'):
             XMAP.append(xi)
             YMAP.append(yi)
             ZMAP.append(Z0)
-    XMAP = np.asfortranarray(XMAP, dtype=dtype)
-    YMAP = np.asfortranarray(YMAP, dtype=dtype)
-    ZMAP = np.asfortranarray(ZMAP, dtype=dtype)
     
-    MAP = np.array([XMAP, YMAP, ZMAP])
-    return MAP.T
+    r_probe = np.transpose([XMAP, YMAP, ZMAP])
+    return r_probe
 
 
-def generate_NF_map_XZ(X0,X1,NX,  Z0,Z1,NZ,  Y0=0, dtype='f'):
+def generate_coord_map_XZ(X0,X1,NX, Z0,Z1,NZ, Y0=0, dtype=None):
+    """wrapper to :func:`generate_NF_map_XZ`"""
+    return generate_NF_map_XZ(X0,X1,NX, Z0,Z1,NZ, Y0, dtype)
+def generate_NF_map_XZ(X0,X1,NX, Z0,Z1,NZ, Y0=0, dtype=None):
     """Generate coordinate list with equidistant positions on a rectangular map in XZ plane
      
     see :func:`generate_NF_map` for doc.
@@ -1877,16 +2092,16 @@ def generate_NF_map_XZ(X0,X1,NX,  Z0,Z1,NZ,  Y0=0, dtype='f'):
             XMAP.append(xi)
             YMAP.append(Y0)
             ZMAP.append(zi)
-    XMAP = np.asfortranarray(XMAP, dtype=dtype)
-    YMAP = np.asfortranarray(YMAP, dtype=dtype)
-    ZMAP = np.asfortranarray(ZMAP, dtype=dtype)
     
-    MAP = np.array([XMAP, YMAP, ZMAP])
-    return MAP.T
+    r_probe = np.transpose([XMAP, YMAP, ZMAP])
+    return r_probe
 
 
-def generate_NF_map_YZ(Y0,Y1,NY,  Z0,Z1,NZ,  X0=0, dtype='f'):
-    """Generate coordinate list with equidistant positions on a rectangular map in YZ plane
+def generate_coord_map_YZ(Y0,Y1,NY, Z0,Z1,NZ, X0=0, dtype=None):
+    """wrapper to :func:`generate_NF_map_YZ`"""
+    return generate_NF_map_YZ(Y0,Y1,NY, Z0,Z1,NZ, X0, dtype)
+def generate_NF_map_YZ(Y0,Y1,NY, Z0,Z1,NZ, X0=0, dtype=None):
+    """Generate coordinate list with equidistant positions on a rectangular map in XZ plane
      
     see :func:`generate_NF_map` for doc.
     """
@@ -1896,17 +2111,51 @@ def generate_NF_map_YZ(Y0,Y1,NY,  Z0,Z1,NZ,  X0=0, dtype='f'):
             XMAP.append(X0)
             YMAP.append(yi)
             ZMAP.append(zi)
-    XMAP = np.asfortranarray(XMAP, dtype=dtype)
-    YMAP = np.asfortranarray(YMAP, dtype=dtype)
-    ZMAP = np.asfortranarray(ZMAP, dtype=dtype)
     
-    MAP = np.array([XMAP, YMAP, ZMAP])
-    return MAP.T
+    r_probe = np.transpose([XMAP, YMAP, ZMAP])
+    return r_probe
+
+
+def generate_coord_map_spherical(r=100000., 
+                                 tetamin=0, tetamax=np.pi/2., Nteta=10, 
+                                 phimin=0, phimax=2*np.pi, Nphi=36):
+    """Generate coordinate list on a spherical screen
+    
+    by default, create coordinates on hemi-spherical upper half-space 
+    
+    Parameters
+    ----------
+    r : float, default: 100000.
+        radius of integration sphere (distance to coordinate origin in nm)
+        
+    tetamin, tetamax : float, float; defaults: 0, np.pi/2
+        minimum and maximum polar angle in radians 
+        (in linear steps from `tetamin` to `tetamax`)
+        
+    phimin, phimax : float, float; defaults: 0, 2*np.pi
+        minimum and maximum azimuth angle in radians, excluding last position
+        (in linear steps from `phimin` to `phimax`)
+        
+    Nteta, Nphi : int, int; defaults: 10, 36
+        number of polar and azimuthal angles on sphere to calculate,
+    
+    Returns
+    -------
+    list of lists :
+        list of 3 lists of X, Y, Z coordinates, to use e.g. in :func:`.linear.nearfield`
+    """
+    tetalist = np.ones((int(Nteta), int(Nphi)))*np.linspace(tetamin, tetamax, int(Nteta))[:,None]
+    philist = np.ones((int(Nteta), int(Nphi)))*np.linspace(phimin, phimax, int(Nphi), endpoint=False)[None,:]
+    xff = (r * np.sin(tetalist) * np.cos(philist)).flatten()
+    yff = (r * np.sin(tetalist) * np.sin(philist)).flatten()
+    zff = (r * np.cos(tetalist)).flatten()
+    r_probe = np.transpose([xff, yff, zff])
+        
+    return r_probe
 
 
 def map_to_grid_XY(MAP, ivalues, NX=-1, NY=-1, map_indices=(0,1)):
-    """
-    Generate XY plottable map e.g. from (XY) nearfield calculation
+    """Generate XY plottable map e.g. from (XY) nearfield calculation
       
     Parameters
     ----------
